from utils import standardize, moralize
import numpy as np
import pandas as pd
import networkx as nx
from causaldag import DAG

def generate_barabasi(ps, m=1, percent_edge_flip=0, random_state=2023):
    Bs, CPDAGs = [], []
    for p in ps:
        G = nx.powerlaw_cluster_graph(n=p, m=m, p=0, seed=random_state)
        B = nx.to_numpy_array(G)
        B = np.tril(B)
        P = np.zeros((p,p))
        for i in range(p):
            P[p-i-1,i] = 1
        B = B @ P
        B = np.rot90(B,3)
        Bs.append(B)
    
    if percent_edge_flip > 0:
        for B in Bs:
            num_reversal = int(len(np.flatnonzero(B)) * percent_edge_flip)
            np.random.seed(random_state)
            indx = np.random.choice(len(np.nonzero(B)[0]), num_reversal)
            row_indx, col_indx = np.nonzero(B)[0][indx], np.nonzero(B)[1][indx]
            
            # reverse edges
            for j in range(num_reversal):
                B[row_indx[j],col_indx[j]] = 0
                B[col_indx[j],row_indx[j]] = 1
                G = nx.from_numpy_array(B, create_using=nx.DiGraph)
                
                # check acyclicity
                if len(list(nx.simple_cycles(G))) != 0:
                    B[row_indx[j],col_indx[j]] = 1
                    B[col_indx[j],row_indx[j]] = 0
    
    for B in Bs:
        CPDAG = DAG.from_dataframe(pd.DataFrame(B)).cpdag().to_amat()[0]
        CPDAGs.append(CPDAG)
               
    return Bs, CPDAGs

def generate_erdos_renyi(ps, exp_num_edges, random_state=2023):
    Bs, CPDAGs = [], []
    for i in range(len(ps)):
        edge_prob = (2*exp_num_edges[i])/(ps[i]*(ps[i]-1))
        G = nx.generators.random_graphs.erdos_renyi_graph(n=ps[i], p=edge_prob, seed=random_state)
        B = nx.to_numpy_matrix(G)
        B = np.tril(B)
        Bs.append(B)
        CPDAG = DAG.from_dataframe(pd.DataFrame(B)).cpdag().to_amat()[0]
        CPDAGs.append(CPDAG)

    return Bs, CPDAGs
    
def generate_data(Bs, lower_weight=0.25, upper_weight=1.0, n_prop_to_p=2, N=20, random_state=2023):
    Xs, Ps = [], []
    for B in Bs:
        p = len(B)
        n = int(n_prop_to_p*p)
        L = np.eye(p) - B
        Omega = L @ L.T
        Omega = np.where(Omega != 0, 1, 0)
        
        np.random.seed(random_state)
        edge_weights = np.random.uniform(low=lower_weight, high=upper_weight, size=(p,p))
        edge_signs = np.random.choice([-1,1], size=(p,p))
        edges = np.multiply(edge_weights, edge_signs)
        Omega = np.multiply(edges, Omega)
        Omega = 0.5*(np.tril(Omega) + np.tril(Omega).T)
        np.fill_diagonal(Omega, 1.2*np.abs(Omega).sum(1))
        
        # set diagonal to 1
        diag_inv = np.diag(1/np.sqrt(np.diag(Omega)))
        Omega = diag_inv @ Omega @ diag_inv
        Cov = np.linalg.inv(Omega)

        for j in range(N):
            rs = np.random.RandomState(random_state)
            X = rs.multivariate_normal(mean=np.zeros(p), cov=Cov, size=n)
            X_std = standardize(X, bias=False)
            
            P = np.eye(p)
            np.random.seed(random_state)
            np.random.shuffle(P.T)
            Ps.append(P)
            X_std = X_std @ P.T
            Xs.append(X_std)

            random_state += 1
            
    return Xs, Ps