import random
from rfd_utils import marginalize_node, random_min, random_max, isclose, iszero, get_submatrix, partial_correlation_threshold
from rfd_utils import get_submatrix
import numpy as np
from scipy.special import comb, binom


# SCORES
def get_degree_score(old_precision, node):
    return np.sum(old_precision[node] != 0) - 1


def get_fill_score(old_precision, node):
    # make mask of only the neighbors of our node
    m = old_precision[node].astype(bool)
    m[node] = 0
    nnbrs = m.sum()
    mask = m[:, None] @ m[None, :]

    # count which neighbors are adjacent
    actual_combos = np.count_nonzero(old_precision * mask)

    possible_combos = nnbrs * (nnbrs - 1)
    score = (possible_combos - actual_combos)/2
    return score


def get_removal_score_full_matrix(old_precision, new_precision, node):
    old_precision_sub = get_submatrix(old_precision, node)
    removed_matrix = ~iszero(old_precision_sub) & iszero(new_precision)
    return removed_matrix.sum()/2


def get_removal_score(old_precision, new_precision, node):
    nbrs = list(set(old_precision[node].nonzero()[0]) - {node})
    if len(nbrs) > old_precision.shape[0]/5:
        return get_removal_score_full_matrix(old_precision, new_precision, node)

    nbrs_new = [j - 1 if j > node else j for j in nbrs]  # reindex
    removed_matrix = iszero(new_precision[np.ix_(nbrs_new, nbrs_new)]) & ~iszero(old_precision[np.ix_(nbrs, nbrs)])
    rscore = removed_matrix.sum()/2
    return rscore


def rfd_plus_step(original_precision, original_precision_thresholded, depth, num_samples, alpha, fill_zeros, verbose=False):
    s = ([], float('inf'), float('inf'), original_precision, original_precision_thresholded)
    current_level_paths = [s]

    while True:
        # terminate if (1) depth exceeded, (2) no nodes remaining, or (3) some path has nonzero removal score
        current_path_length = len(current_level_paths[0][0])
        if verbose: print(f"[rfd_plus_step] Current path is of length {current_path_length}")

        min_removed_score = min(current_level_paths, key=lambda x: x[1])[1]
        min_removed_paths = [path for path in current_level_paths if path[1] == min_removed_score]

        no_nodes_remaining = len(original_precision) - current_path_length == 0
        if depth == 0 or no_nodes_remaining or min_removed_score < 0:  # lines 7-9, and 10
            if verbose and min_removed_score < 0: print(f"[rfd_plus_step] Picking path with removal score {min_removed_score}")
            min_degree = min([path[2] for path in min_removed_paths])
            min_degree_paths = [path for path in min_removed_paths if path[2] == min_degree]
            random_ix = random.randint(0, len(min_degree_paths)-1)
            selected_path = min_degree_paths[random_ix]

            path_precision = selected_path[-2]
            path_precision_thresholded = selected_path[-1]
            return selected_path[0], path_precision, path_precision_thresholded

        next_level_paths = []
        depth -= 1
        if verbose: print(f"Searching for new paths")
        best_min_removal_score = 0
        for path, _, _, path_precision, path_precision_thresholded in current_level_paths:
            nodes = list(range(len(path_precision)))
            nodes2new_precision = {
                node: marginalize_node(path_precision, node, num_samples=num_samples, alpha=alpha, fill_zeros=fill_zeros)
                for node in nodes
            }
            nodes2_new_precision_thresholded = {
                node: partial_correlation_threshold(new_precision, num_samples, alpha)
                for node, new_precision in nodes2new_precision.items()
            }
            nodes2removal_score = {
                node: -get_removal_score(path_precision_thresholded, new_precision_thresholded, node)
                for node, new_precision_thresholded in nodes2_new_precision_thresholded.items()
            }

            min_removal_score = min((score for _, score in nodes2removal_score.items()))
            if best_min_removal_score < min_removal_score:  # skip if a better removal score has already been found
                next_level_nodes = []
            elif min_removal_score <= best_min_removal_score:  # if this matches the best removal score or is better, append
                best_min_removal_score = min_removal_score
                next_level_nodes = [node for node, score in nodes2removal_score.items() if score == min_removal_score]
            else:  # otherwise, best removal score so far is 0, so expand all things with min_fill_score
                nodes2fill_score = {node: get_fill_score(path_precision, node) for node in nodes}
                min_fill_score = min((score for _, score in nodes2fill_score.items()))
                next_level_nodes = [node for node, score in nodes2fill_score.items() if score == min_fill_score]

            for node in next_level_nodes:
                degree_score = get_degree_score(path_precision, node)
                new_precision = nodes2new_precision[node]
                new_precision_thresholded = nodes2_new_precision_thresholded[node]
                next_level_paths.append((path + [node], min_removal_score, degree_score, new_precision, new_precision_thresholded))

        current_level_paths = next_level_paths


def greedy_algorithm_wrapper(
        original_precision,
        node_selector,
        num_samples=None,
        alpha=None,
        verbose=False,
        fill_zeros=False,
):
    remaining_nodes = list(range(original_precision.shape[0]))
    permutation = []

    current_precision = original_precision
    current_precision_thresholded = partial_correlation_threshold(current_precision, num_samples, alpha)
    while remaining_nodes:
        if verbose:
            print(f"{len(remaining_nodes)} nodes remaining")

        next_nodes, new_precision, new_precision_thresholded = node_selector(
            current_precision,
            current_precision_thresholded,
            num_samples=num_samples,
            alpha=alpha,
            fill_zeros=fill_zeros
        )

        if new_precision is None:
            assert len(next_nodes) == 1
            new_precision = marginalize_node(current_precision, next_nodes[0], num_samples=num_samples, alpha=alpha, fill_zeros=fill_zeros)
            new_precision_thresholded = partial_correlation_threshold(new_precision, num_samples, alpha)
        for node in next_nodes:
            permutation.append(remaining_nodes[node])

            del remaining_nodes[node]
            current_precision_thresholded = get_submatrix(current_precision_thresholded, node)
        current_precision = new_precision
        current_precision_thresholded = new_precision_thresholded

    assert set(permutation) == set(range(original_precision.shape[0]))
    return list(reversed(permutation))


def min_degree_selector(current_precision, current_precision_thresholded, num_samples=None, alpha=None, fill_zeros=False):
    nodes = list(range(current_precision.shape[0]))
    node2degree_score = {node: get_degree_score(current_precision, node) for node in nodes}
    return [random_min(node2degree_score)], None, None


def minimum_degree_algorithm(precision, num_samples=None, alpha=None, verbose=False, fill_zeros=False):
    return greedy_algorithm_wrapper(precision, min_degree_selector, num_samples, alpha, verbose=verbose, fill_zeros=fill_zeros)


def min_fill_selector(current_precision, current_precision_thresholded, num_samples=None, alpha=None, fill_zeros=False):
    nodes = list(range(current_precision.shape[0]))
    node2fill_score = {node: get_fill_score(current_precision, node) for node in nodes}
    return [random_min(node2fill_score)], None, None


def min_fill_algorithm(precision, num_samples=None, alpha=None, verbose=False, fill_zeros=False):
    return greedy_algorithm_wrapper(precision, min_fill_selector, num_samples, alpha, verbose=verbose, fill_zeros=fill_zeros)


def max_remove_selector(current_precision, current_precision_thresholded, num_samples=None, alpha=None, fill_zeros=False):
    nodes = list(range(current_precision.shape[0]))
    node2new_precision = {
        node: marginalize_node(current_precision, node, num_samples, alpha, fill_zeros=fill_zeros)
        for node in nodes
    }
    node2thresholded_new_precision = {
        node: partial_correlation_threshold(new_precision, num_samples, alpha)
        for node, new_precision in node2new_precision.items()
    }
    node2removal_score = {
        node: get_removal_score(current_precision_thresholded, new_precision_thresholded, node)
        for node, new_precision_thresholded in node2thresholded_new_precision.items()
    }
    selected_node = random_max(node2removal_score)
    return [selected_node], node2new_precision[selected_node], node2thresholded_new_precision[selected_node]


def max_remove_algorithm(precision, num_samples=None, alpha=None, verbose=False, fill_zeros=False):
    return greedy_algorithm_wrapper(precision, max_remove_selector, num_samples, alpha, verbose=verbose, fill_zeros=fill_zeros)


def rfd_algorithm(precision, num_samples=None, alpha=None, depth=1, verbose=False, fill_zeros=False):

    def rfd_selector(current_precision,  current_precision_thresholded, num_samples=num_samples, alpha=alpha, fill_zeros=False):
        return rfd_plus_step(current_precision, current_precision_thresholded, depth=depth, num_samples=num_samples, alpha=alpha, fill_zeros=fill_zeros)

    return greedy_algorithm_wrapper(precision, rfd_selector, num_samples, alpha, verbose=verbose, fill_zeros=fill_zeros)


if __name__ == '__main__':
    import causaldag as cd
    from causaldag.utils.ci_tests import MemoizedCI_Tester, gauss_ci_test, dsep_test


    # minimum_degree_algorithm(precision, num_samples=nsamples, alpha=alpha, verbose=True)
    # min_fill_algorithm(precision, num_samples=nsamples, alpha=alpha, verbose=True)
    # max_remove_algorithm(precision, num_samples=nsamples, alpha=alpha, verbose=True)
    noisy = False
    for _ in range(10):
        d = cd.rand.directed_erdos(5, 1/2)
        g = cd.rand.rand_weights(d)
        nsamples = 100
        alpha = .01
        samples = g.sample(nsamples)
        suffstat = cd.gauss_ci_suffstat(samples)
        if noisy:
            precision = suffstat['P']
            perm, d_est1 = rfd_algorithm(precision, num_samples=nsamples, alpha=alpha, depth=2, construct_imap=True)
            # perm, d_est1 = rfd_algorithm(precision, num_samples=nsamples, alpha=alpha, depth=2, construct_imap=True)
            # print(d_est1.is_topological(perm))

            ci_tester = MemoizedCI_Tester(gauss_ci_test, suffstat, alpha=alpha)
            d_est2 = cd.perm2dag(perm, ci_tester)
        else:
            precision = g.precision
            perm, d_est1 = rfd_algorithm(precision, depth=2, construct_imap=True)
            # perm, d_est1 = rfd_algorithm(precision, num_samples=nsamples, alpha=alpha, depth=2, construct_imap=True)
            # print(d_est1.is_topological(perm))

            ci_tester = MemoizedCI_Tester(dsep_test, d)
            d_est2 = cd.perm2dag(perm, ci_tester)
        # print(d.shd(d_est1))
        # print(d.shd(d_est2))
        # print(d_est1.shd(d_est2))
        if not d_est1.arcs >= d_est2.arcs:
            print("Error")
            break
        print(d_est1.arcs >= d_est2.arcs)

