import numpy as np
from scipy import special
import itertools as itr
import cvxpy as cp
import random
import ipdb
from graphical_models import DAG


def get_submatrix(m, missing_node):
    new_p = m.shape[0] - 1
    c = np.empty((new_p, new_p))
    c[:missing_node, :missing_node] = m[:missing_node, :missing_node]
    c[missing_node:, :missing_node] = m[(missing_node+1):, :missing_node]
    c[:missing_node, missing_node:] = m[:missing_node, (missing_node+1):]
    c[missing_node:, missing_node:] = m[(missing_node+1):, (missing_node+1):]
    return c


def get_subcolumn(m, missing_node):
    """
    Return the column `missing_node`, without the row `missing_node`. i.e., from a p*p matrix, a (p-1)-length vector
    is returned
    """
    new_p = m.shape[0] - 1
    c = np.empty(new_p)
    c[:missing_node] = m[:missing_node, missing_node]
    c[missing_node:] = m[(missing_node+1):, missing_node]
    return c


def iszero(a, atol=1e-8):
    return abs(a) < atol


def isclose(a, b, atol=1e-8):
    return abs(a - b) < atol


def random_min(d: dict):
    min_val = min(d.values())
    min_keys = [k for k, v in d.items() if v == min_val]
    return random.choice(min_keys)


def random_max(d: dict):
    min_val = max(d.values())
    min_keys = [k for k, v in d.items() if v == min_val]
    return random.choice(min_keys)


def mle(amat, S):
    zero_ix = zip(*np.where(amat == 0))
    zero_ix = [(i, j) for i, j in zero_ix if i != j]
    X = cp.Variable(amat.shape, PSD = True)
    constraints = [X[ix] == 0 for ix in zero_ix]
    prob = cp.Problem(cp.Maximize(cp.log_det(X) - cp.sum(cp.multiply(X, S))),
                constraints)
    prob.solve(solver='SCS', verbose=False)
    if X.value is None:
        print(prob.solution.status)
        raise ValueError
    return X.value


def partial_correlation_threshold(precision, n=None, alpha=None):
    if n is None:
        return precision

    assert(len(precision.shape) == 2)
    r = precision/np.sqrt(np.diagonal(precision))/np.sqrt(np.diagonal(precision))[:, None]
    p = r.shape[0]
    n_cond = p - 2

    # note: log1p(2r/(1-r)) = log((1+r)/(1-r)) but is more numerically stable for r near 0
    # r = 1 causes warnings but gives the correct answer
    with np.errstate(divide='ignore', invalid='ignore'):
        statistic = np.sqrt(n - n_cond - 3) * abs(.5 * np.log1p(2*r/(1 - r)))

    p_values = 1 - .5*(1 + special.erf(statistic/np.sqrt(2)))

    zero_ixs = p_values > alpha
    np.fill_diagonal(zero_ixs, False)
    r[zero_ixs] = 0

    return r * np.sqrt(np.diagonal(precision))*np.sqrt(np.diagonal(precision))[:, None]


def marginalize_node(precision, node, num_samples=None, alpha=None, fill_zeros=False):
    if not fill_zeros or num_samples is None:
        precision11 = get_submatrix(precision, node)
        precision22 = precision[node, node]
        precision21 = get_subcolumn(precision, node)  # this is faster for large arrays, e.g. p > 50.
        return precision11 - np.outer(precision21 * np.array(1/precision22), precision21)
    else:
        return marginalize_node_with_thresholding(precision, node, num_samples, alpha)


def marginalize_node_with_thresholding(precision, node, num_samples, alpha):
    n = precision.shape[0]
    threshold_precision = partial_correlation_threshold(precision,num_samples,alpha)
    wo_node = list(range(node)) + list(range(node+1, precision.shape[0]))

    precision11 = get_submatrix(precision, node)
    precision21 = precision[node][wo_node]
    precision22 = precision[node][node]

    th_precision21 = threshold_precision[node][wo_node]
    th_precision22 = threshold_precision[node][node]

    updated_precision = np.zeros((n,n))
    updated_precision[np.ix_(wo_node,wo_node)] =  precision11 - 1/precision22 * np.outer(precision21,precision21) + 1/th_precision22 * np.outer(th_precision21,th_precision21)
    updated_precision[node] = threshold_precision[node]
    updated_precision.T[node] = threshold_precision[node]

    if abs(precision22) < 1e-10:
        raise ValueError(f"precision22: {precision22}")

    if abs(th_precision22) < 1e-10:
        raise ValueError(f"th_precision22: {th_precision22}")

    eigvals,eigvecs = np.linalg.eig(updated_precision)
    if ((eigvals < 0 ).any()):
        updated_precision = eigvecs * np.maximum(eigvals,0) @ eigvecs.T

    assert precision.shape == updated_precision.shape

    return marginalize_node(updated_precision,node)

def perm2dag_precision(perm, precision, alpha=.01, num_samples=None):
    perm = np.array(perm)
    current_precision = precision.copy()
    current_precision[:, :] = current_precision[perm, :]
    current_precision[:, :] = current_precision[:, perm]
    nnodes = precision.shape[0]
    arcs = set()

    # iterate through the permutation in reverse order
    for node in range(nnodes-1, -1, -1):
        if num_samples is not None:
            current_precision_thresholded = partial_correlation_threshold(current_precision, n=num_samples, alpha=alpha)
            parents = np.nonzero(current_precision_thresholded[-1])[0]
        else:
            parents = np.nonzero(~iszero(current_precision[-1]))[0]
        label = perm[node]
        new_arcs = set(itr.product(perm[parents], [label])) - {(label, label)}
        arcs.update(new_arcs)

        # marginalize out the last node
        current_precision = current_precision[:-1, :-1] - current_precision[-1, -1]**-1 * np.outer(current_precision[:-1, -1], current_precision[:-1, -1])

    return DAG(nodes=set(perm), arcs=arcs)


if __name__ == '__main__':
    import causaldag as cd

    nnodes = 10
    d = cd.rand.directed_erdos(nnodes, 1/2)
    g = cd.rand.rand_weights(d)
    samples = g.sample(100)
    cov = np.cov(samples, rowvar=False)
    prec = np.linalg.inv(cov)

    for node in list(range(nnodes)):
        cov_marg = np.delete(np.delete(cov, node, axis=0), node, axis=1)
        prec_marg = np.linalg.inv(cov_marg)
        prec_marg2 = marginalize_node(prec, node)
        print(np.max(np.abs(prec_marg - prec_marg2)))

