import numpy as np
import sklearn
from sklearn.metrics import confusion_matrix
import networkx as nx

def standardize(X, bias=False):
    """
    Standardize dataset.
    """
    X_std = X.copy()
    X_std -= X_std.mean(axis=0)
    if bias:
        X_std /= X_std.std(axis=0)
    else:
        X_std /= X_std.std(axis=0, ddof=1)
    
    return X_std

def moralize(B):
    G = nx.from_numpy_matrix(B, create_using=nx.DiGraph)
    moral_graph = nx.to_numpy_array(nx.moral_graph(G))
    
    return moral_graph

def compute_shd(CPDAG, CPDAG_hat):
    """
    Compute SHD.
    """
    CPDAG_copy = CPDAG.copy()
    CPDAG_hat_copy = CPDAG_hat.copy()
    
    # skeleton
    Skel = np.tril(np.where(CPDAG_copy + CPDAG_copy.T != 0, 1, 0))
    Skel_hat = np.tril(np.where(CPDAG_hat_copy + CPDAG_hat_copy.T != 0, 1, 0))
    
    # set undirected edges to -1
    CPDAG_copy[np.where(CPDAG_copy+CPDAG_copy.T==2)] = -1
    CPDAG_hat_copy[np.where(CPDAG_hat_copy+CPDAG_hat_copy.T==2)] = -1
    
    # number of true edges
    T = len(np.flatnonzero(Skel))
    
    # number of estimated edges
    P = len(np.flatnonzero(Skel_hat))
    
    # number of true positives
    TP_dir = len(np.intersect1d(np.flatnonzero(CPDAG_copy==1), np.flatnonzero(CPDAG_hat_copy==1), assume_unique=True))
    TP_undir = .5*len(np.intersect1d(np.flatnonzero(CPDAG_copy==-1), np.flatnonzero(CPDAG_hat_copy==-1), assume_unique=True))
    TP = TP_dir + TP_undir
    
    # number of false positives
    FP = len(np.setdiff1d(np.flatnonzero(Skel_hat), np.flatnonzero(Skel), assume_unique=True))
    
    # number of reversed edges (including undirected-directed edges)
    R = P - TP - FP

    # number of missing edges
    M = len(np.setdiff1d(np.flatnonzero(Skel), np.flatnonzero(Skel_hat), assume_unique=True))
    
    # structural hamming distance
    SHD = FP + M + R
    
    return SHD

def compute_confusion_matrix(CPDAG, CPDAG_hat, diag=False):
    """
    Compute entries of confusion matrix with FDR.
    If diag=False, it excludes diagonal elements before computing.
    """
    CPDAG_vec = CPDAG.ravel()
    CPDAG_hat_vec = CPDAG_hat.ravel()
    
    if not diag:
        p = len(CPDAG)
        indx = np.arange(0, p**2, p+1)
        CPDAG_vec = np.where(CPDAG_vec != 0, 1, 0)
        CPDAG_vec = np.delete(CPDAG_vec, indx)
        CPDAG_hat_vec = np.where(CPDAG_hat_vec != 0, 1, 0)
        CPDAG_hat_vec = np.delete(CPDAG_hat_vec, indx)
    TN, FP, FN, TP = confusion_matrix(CPDAG_vec, CPDAG_hat_vec).ravel()
    if TP + FP != 0:
        FDR = FP/(TP+FP)
    else:
        FDR = np.nan
    
    return TP, FP, TN, FN, FDR

def mcc(CPDAG, CPDAG_hat, diag=False):
    """
    Compute Matthew's Correlation Coefficient (MCC) between two matrices.
    If diag=False, it excludes diagonal elements before computing.
    """
    CPDAG_vec = CPDAG.ravel()
    CPDAG_hat_vec = CPDAG_hat.ravel()
    
    if not diag:
        p = len(CPDAG)
        indx = np.arange(0, p**2, p+1)
        CPDAG_vec = np.where(CPDAG_vec != 0, 1, 0)
        CPDAG_vec = np.delete(CPDAG_vec, indx)
        CPDAG_hat_vec = np.where(CPDAG_hat_vec != 0, 1, 0)
        CPDAG_hat_vec = np.delete(CPDAG_hat_vec, indx)
    
    return sklearn.metrics.matthews_corrcoef(CPDAG_vec, CPDAG_hat_vec)